"""Top-level package for fmu-datamodels."""

from fmu.datamodels._schema_base import SchemaBase

from .common import (
    Access,
    Asset,
    CoordinateSystem,
    CountryItem,
    DiscoveryItem,
    FieldItem,
    Masterdata,
    OperatingSystem,
    Smda,
    Ssdl,
    SsdlAccess,
    StratigraphicColumn,
    SystemInformation,
    Tracklog,
    TracklogEvent,
    TracklogSource,
    User,
    Version,
)
from .fmu_results import FmuResults, FmuResultsSchema
from .standard_results import (
    ErtDistribution,
    ErtParameterMetadata,
    ErtParametersResult,
    ErtParametersSchema,
    FieldOutlineResult,
    FieldOutlineSchema,
    FluidContactOutlineResult,
    FluidContactOutlineSchema,
    InplaceVolumesResult,
    InplaceVolumesSchema,
    SimulatorFipregionsMappingResult,
    SimulatorFipregionsMappingSchema,
    StructureDepthFaultLinesResult,
    StructureDepthFaultLinesSchema,
)

try:
    from .version import version

    __version__ = version
except ImportError:
    __version__ = "0.0.0"

__all__ = [
    "Access",
    "SsdlAccess",
    "Asset",
    "Ssdl",
    "Masterdata",
    "Smda",
    "StratigraphicColumn",
    "CoordinateSystem",
    "CountryItem",
    "FieldItem",
    "DiscoveryItem",
    "Tracklog",
    "OperatingSystem",
    "SystemInformation",
    "TracklogEvent",
    "TracklogSource",
    "User",
    "Version",
    "ErtDistribution",
    "ErtParameterMetadata",
    "ErtParametersResult",
    "ErtParametersSchema",
    "FmuResults",
    "FmuResultsSchema",
    "FieldOutlineResult",
    "FieldOutlineSchema",
    "FluidContactOutlineResult",
    "FluidContactOutlineSchema",
    "InplaceVolumesResult",
    "InplaceVolumesSchema",
    "SimulatorFipregionsMappingResult",
    "SimulatorFipregionsMappingSchema",
    "StructureDepthFaultLinesResult",
    "StructureDepthFaultLinesSchema",
]

schemas: list[type[SchemaBase]] = [
    FmuResultsSchema,
    ErtParametersSchema,
    FieldOutlineSchema,
    FluidContactOutlineSchema,
    InplaceVolumesSchema,
    SimulatorFipregionsMappingSchema,
    StructureDepthFaultLinesSchema,
]



import datetime
from uuid import UUID

import pytest
from pydantic import BaseModel

from fmu.datamodels.common.access import (
    Access,
    Asset,
    Ssdl,
    SsdlAccess,
)
from fmu.datamodels.common.enums import Classification, TrackLogEventType
from fmu.datamodels.common.masterdata import (
    CoordinateSystem,
    CountryItem,
    DiscoveryItem,
    FieldItem,
    Masterdata,
    Smda,
    StratigraphicColumn,
)
from fmu.datamodels.common.tracklog import (
    OperatingSystem,
    SystemInformation,
    Tracklog,
    TracklogEvent,
    User,
    Version,
)
from fmu.datamodels.fmu_results import enums
from fmu.datamodels.fmu_results.data import (
    AnyData,
    BoundingBox2D,
    BoundingBox3D,
    FieldOutline,
    FieldRegion,
    FluidContact,
    Layer,
    Seismic,
    SmdaEntity,
    Time,
    Timestamp,
)
from fmu.datamodels.fmu_results.fields import (
    FMU,
    Case,
    Context,
    Display,
    File,
    FMUBase,
    Model,
)
from fmu.datamodels.fmu_results.fmu_results import (
    CaseMetadata,
    FmuResults,
    FmuResultsSchema,
    ObjectMetadata,
)
from fmu.datamodels.fmu_results.global_configuration import (
    Stratigraphy,
    StratigraphyElement,
)
from fmu.datamodels.fmu_results.specification import (
    PolygonsSpecification,
    Statistics,
    SurfaceSpecification,
    TableSpecification,
)
from fmu.datamodels.types import MD5HashStr
from tests.utils import _get_nested_pydantic_models


@pytest.fixture(scope="session")
def pydantic_models_from_root() -> set[type[BaseModel]]:
    """Return all nested pydantic models from FmuResults and downwards"""
    return _get_nested_pydantic_models(FmuResults)


@pytest.fixture(name="testdata_stratigraphy", scope="function")
def fixture_testdata_stratigraphy() -> Stratigraphy:
    """
    Return a dict of StratigraphyElement instances.
    """
    return Stratigraphy(
        root={
            "TopStratUnit1": StratigraphyElement(
                name="Stratigraphic Unit 1",
                stratigraphic=True,
                alias=["TopSU1", "TopLayer1"],
            ),
            "TopStratUnit2": StratigraphyElement(
                name="Stratigraphic Unit 2",
                stratigraphic=True,
                alias=["TopSU2", "TopLayer2"],
            ),
            "TopStratUnit3": StratigraphyElement(
                name="Stratigraphic Unit 3",
                stratigraphic=True,
                alias=["TopSU3", "TopLayer3"],
            ),
        }
    )


def _generate_metadata_base() -> dict:
    """Generates an example of the metadata base in FMU results"""

    masterdata = Masterdata.model_construct(
        smda=Smda.model_construct(
            coordinate_system=CoordinateSystem.model_construct(
                identifier="ST_WGS84_UTM37N_P32637",
                uuid=UUID("00000000-0000-0000-0000-000000000000"),
            ),
            country=[
                CountryItem.model_construct(
                    identifier="SomeDiscovery",
                    uuid=UUID("00000000-0000-0000-0000-000000000000"),
                )
            ],
            discovery=[
                DiscoveryItem.model_construct(
                    short_identifier="Norway",
                    uuid=UUID("00000000-0000-0000-0000-000000000000"),
                )
            ],
            field=[
                FieldItem.model_construct(
                    identifier="OseFax",
                    uuid=UUID("00000000-0000-0000-0000-000000000000"),
                )
            ],
            stratigraphic_column=StratigraphicColumn.model_construct(
                identifier="DROGON_2020",
                uuid=UUID("00000000-0000-0000-0000-000000000000"),
            ),
            smda_entity=SmdaEntity.model_construct(
                identifier="Viking Gp. Top",
                uuid=UUID("00000000-0000-0000-0000-000000000000"),
            ),
        )
    )

    tracklog = Tracklog.model_construct(
        [
            TracklogEvent.model_construct(
                datetime=datetime.datetime.now(datetime.UTC),
                event=TrackLogEventType.created,
                user=User(id="user"),
                sysinfo=SystemInformation.model_construct(
                    fmu_dataio=Version.model_construct(version="dummy_version"),
                    komodo=Version.model_construct(version="dummy_version"),
                    operating_system=OperatingSystem(
                        hostname="dummy_hostname",
                        operating_system="dummy_os",
                        release="dummy_release",
                        system="dummy_system",
                        version="dummy_version",
                    ),
                ),
            )
        ]
    )

    return {
        "masterdata": masterdata,
        "tracklog": tracklog,
        "source": "fmu",
        "version": FmuResultsSchema.VERSION,
        "schema": "https://schema_url.com",
    }


def _generate_fmu_base() -> FMUBase:
    """Generates an example of the FMU base in FMU results"""

    return FMUBase.model_construct(
        case=Case.model_construct(
            name="MyCaseName",
            user=User(id="user"),
            uuid=UUID("00000000-0000-0000-0000-000000000000"),
            description=None,
        ),
        model=Model.model_construct(
            description=None, name="Drogon", revision="21.0.0.dev"
        ),
    )


def _generate_object_metadata_base() -> dict:
    """Generate a base for the ObjectMetadata object, without the 'data' field set"""

    metadata_base = _generate_metadata_base()
    fmu_base = _generate_fmu_base()

    fmu = FMU.model_construct(
        case=fmu_base.case,
        model=fmu_base.model,
        fmu_base=fmu_base,
        context=Context(stage=enums.FMUContext.realization),
        ensemble=None,
        iteration=None,
        workflow=None,
        aggregation=None,
        realization=None,
        entity=None,
        ert=None,
    )

    access = SsdlAccess.model_construct(
        asset=Asset(name="test"),
        ssdl=Ssdl.model_construct(
            access_level=Classification.internal, rep_include=False
        ),
    )

    file = File.model_construct(
        absolute_path="/some/absolute/path/",
        relative_path="realization-0/iter-0/share/results/maps/some_file_name",
        runpath_relative_path=None,
        checksum_md5=MD5HashStr("ab0fb42049138871749c6c367d2d094f"),
        size_bytes=None,
    )

    object_metadata_dict = {
        "fmu": fmu,
        "access": access,
        "file": file,
        "display": Display(name="VIKING GP. Top"),
    }

    object_metadata_dict.update(metadata_base)
    return object_metadata_dict


@pytest.fixture(scope="function")
def case_metadata() -> dict:
    """Generate an example of valid case metadata"""

    metadata_base = _generate_metadata_base()
    fmu_base = _generate_fmu_base()

    access = Access.model_construct(
        asset=Asset(name="test"), classification=Classification.internal
    )

    case_metadata_dict = {
        "class": enums.FMUResultsMetadataClass.case,
        "fmu": fmu_base,
        "access": access,
    }

    case_metadata_dict.update(metadata_base)
    return CaseMetadata.model_validate(case_metadata_dict).model_dump(
        mode="json", exclude_none=True, by_alias=True
    )


@pytest.fixture(scope="function")
def fluid_contact_metadata() -> dict:
    """Generate an example of valid fluid contact metadata"""

    object_metadata_dict = _generate_object_metadata_base()

    data = AnyData.model_validate(
        {
            "content": enums.Content.fluid_contact,
            "standard_result": None,
            "name": "VIKING GP. Top",
            "alias": None,
            "tagname": None,
            "stratigraphic": False,
            "description": None,
            "geometry": None,
            "bbox": BoundingBox3D(
                xmin=456063.6875,
                xmax=467483.6875,
                ymin=5926551.0,
                ymax=5939431.0,
                zmax=1933.5001220703125,
                zmin=1554.5228271484375,
            ),
            "format": enums.FileFormat.irap_binary,
            "grid_model": None,
            "is_observation": False,
            "is_prediction": False,
            "layout": None,
            "offset": 0.6,
            "spec": SurfaceSpecification(
                nrow=5,
                ncol=5,
                rotation=0.1,
                undef=0.0,
                xinc=0.1,
                yinc=0.1,
                xori=0.1,
                yflip=enums.AxisOrientation.normal,
                yori=0.1,
                value_statistics=Statistics(min=-1.0, max=1.0, mean=0.0, std=0.5),
            ),
            "time": None,
            "undef_is_zero": None,
            "unit": "m",
            "vertical_domain": "depth",
            "domain_reference": None,
            "table_index": None,
            "base": None,
            "top": None,
            "fluid_contact": FluidContact(
                contact=enums.FluidContactType.goc,
                truncated=False,
            ),
        }
    )

    object_metadata_dict["class"] = enums.ObjectMetadataClass.surface
    object_metadata_dict["data"] = data

    return ObjectMetadata.model_validate(object_metadata_dict).model_dump(
        mode="json", exclude_none=True, by_alias=True
    )


@pytest.fixture(scope="function")
def field_outline_metadata() -> dict:
    """Generate an example of valid field outline metadata"""

    object_metadata_dict = _generate_object_metadata_base()

    data = AnyData.model_validate(
        {
            "content": enums.Content.field_outline,
            "standard_result": None,
            "name": "VOLANTIS GP. Base",
            "alias": None,
            "tagname": None,
            "stratigraphic": False,
            "description": None,
            "geometry": None,
            "bbox": None,
            "format": enums.FileFormat.csv_xtgeo,
            "grid_model": None,
            "is_observation": False,
            "is_prediction": False,
            "layout": None,
            "offset": 0.6,
            "spec": PolygonsSpecification(
                npolys=2,
                columns=["X_UTME", "Y_UTMN", "Z_TVDSS", "POLY_ID"],
                num_columns=4,
                num_rows=200,
                size=800,
            ),
            "time": None,
            "undef_is_zero": None,
            "unit": "m",
            "vertical_domain": "depth",
            "domain_reference": None,
            "table_index": None,
            "base": None,
            "top": None,
            "field_outline": FieldOutline(
                contact="goc",
            ),
        }
    )

    object_metadata_dict["class"] = enums.ObjectMetadataClass.polygons
    object_metadata_dict["data"] = data

    return ObjectMetadata.model_validate(object_metadata_dict).model_dump(
        mode="json", exclude_none=True, by_alias=True
    )


@pytest.fixture(scope="function")
def field_region_metadata() -> dict:
    """Generate valid field region metadata"""

    object_metadata_dict = _generate_object_metadata_base()

    data = AnyData.model_validate(
        {
            "content": enums.Content.field_region,
            "standard_result": None,
            "name": "VOLANTIS GP. Base",
            "alias": None,
            "tagname": None,
            "stratigraphic": False,
            "description": None,
            "geometry": None,
            "bbox": None,
            "format": enums.FileFormat.csv_xtgeo,
            "grid_model": None,
            "is_observation": False,
            "is_prediction": False,
            "layout": None,
            "offset": 0.6,
            "spec": PolygonsSpecification(
                npolys=2,
                columns=["X_UTME", "Y_UTMN", "Z_TVDSS", "POLY_ID"],
                num_columns=4,
                num_rows=200,
                size=800,
            ),
            "time": None,
            "undef_is_zero": None,
            "unit": "m",
            "vertical_domain": None,
            "domain_reference": None,
            "table_index": None,
            "base": None,
            "top": None,
            "field_region": FieldRegion(
                id=99,
            ),
        }
    )

    object_metadata_dict["class"] = enums.ObjectMetadataClass.polygons
    object_metadata_dict["data"] = data

    return ObjectMetadata.model_validate(object_metadata_dict).model_dump(
        mode="json", exclude_none=True, by_alias=True
    )


@pytest.fixture(scope="function")
def seismic_metadata() -> dict:
    """Generate valid seismic metadata"""

    object_metadata_dict = _generate_object_metadata_base()

    data = AnyData.model_validate(
        {
            "content": enums.Content.seismic,
            "standard_result": None,
            "name": "VIKING GP. Top",
            "alias": None,
            "tagname": None,
            "stratigraphic": True,
            "description": None,
            "geometry": None,
            "bbox": BoundingBox2D(
                xmin=456063.6875, xmax=467483.6875, ymin=5926551.0, ymax=5939431.0
            ),
            "format": enums.FileFormat.irap_binary,
            "grid_model": None,
            "is_observation": False,
            "is_prediction": True,
            "layout": enums.Layout.regular,
            "offset": 0.0,
            "spec": SurfaceSpecification(
                ncol=20,
                nrow=100,
                rotation=0.0,
                undef=0.0,
                xinc=0.1,
                yinc=0.1,
                xori=456063.6875,
                yori=5926551.0,
                yflip=enums.AxisOrientation.normal,
                value_statistics=Statistics(min=-1.0, max=1.0, mean=0.0, std=0.5),
            ),
            "time": Time(
                t0=Timestamp(label="base", value=datetime.datetime.now(datetime.UTC)),
                t1=Timestamp(
                    label="monitor", value=datetime.datetime.now(datetime.UTC)
                ),
            ),
            "undef_is_zero": False,
            "unit": "m",
            "vertical_domain": "depth",
            "domain_reference": "msl",
            "table_index": None,
            "base": Layer(name="volantis_gp_base", offset=1.0, stratigraphic=True),
            "top": Layer(name="volantis_gp_top", offset=2.0, stratigraphic=True),
            "seismic": Seismic(attribute="amplitude", calculation="mean"),
        }
    )

    object_metadata_dict["class"] = enums.ObjectMetadataClass.surface
    object_metadata_dict["data"] = data

    return ObjectMetadata.model_validate(object_metadata_dict).model_dump(
        mode="json", exclude_none=True, by_alias=True
    )


@pytest.fixture(scope="function")
def volumes_metadata() -> dict:
    """Generate valid volumes metadata"""

    object_metadata_dict = _generate_object_metadata_base()

    data = AnyData.model_validate(
        {
            "content": enums.Content.volumes,
            "standard_result": None,
            "name": "geogrid",
            "alias": None,
            "tagname": None,
            "stratigraphic": False,
            "description": None,
            "geometry": None,
            "bbox": None,
            "format": enums.FileFormat.csv,
            "grid_model": None,
            "is_observation": False,
            "is_prediction": True,
            "layout": enums.Layout.table,
            "offset": 0.0,
            "spec": TableSpecification(
                columns=["ZONE", "REGION", "FLUID", "BULK_OIL"],
                num_columns=4,
                num_rows=10,
                size=40,
            ),
            "undef_is_zero": False,
            "unit": "m",
            "vertical_domain": "depth",
            "domain_reference": "msl",
            "table_index": ["FLUID", "ZONE", "REGION"],
            "base": None,
            "top": None,
        }
    )

    object_metadata_dict["class"] = enums.ObjectMetadataClass.table
    object_metadata_dict["data"] = data

    return ObjectMetadata.model_validate(object_metadata_dict).model_dump(
        mode="json", exclude_none=True, by_alias=True
    )


@pytest.fixture(scope="function")
def property_metadata() -> dict:
    """Generate valid property metadata"""

    object_metadata_dict = _generate_object_metadata_base()

    data = AnyData.model_validate(
        {
            "content": enums.Content.property,
            "property": {"attribute": "porosity", "is_discrete": False},
            "name": "phit",
            "stratigraphic": False,
            "format": enums.FileFormat.roff,
            "is_observation": False,
            "is_prediction": True,
            "layout": enums.Layout.cornerpoint,
            "offset": 0.0,
            "undef_is_zero": False,
            "unit": "m",
            "vertical_domain": "depth",
            "domain_reference": "msl",
        }
    )

    object_metadata_dict["class"] = enums.ObjectMetadataClass.cpgrid_property
    object_metadata_dict["data"] = data

    return ObjectMetadata.model_validate(object_metadata_dict).model_dump(
        mode="json", exclude_none=True, by_alias=True
    )



import pytest
from pydantic import TypeAdapter

from fmu.datamodels.fmu_results.standard_result import AnyStandardResult
from fmu.datamodels.standard_results.enums import StandardResultName

StandardResultAdapter = TypeAdapter(AnyStandardResult)


@pytest.mark.parametrize("name", list(StandardResultName))
def test_all_standard_results_in_anystandardresult(name: StandardResultName) -> None:
    """Every StandardResultName resolves to a valid AnyStandardResult."""
    standard_result = StandardResultAdapter.validate_python({"name": name})
    assert isinstance(standard_result, AnyStandardResult)



from typing import Annotated, TypeAlias

from pydantic import Field

VersionStr: TypeAlias = Annotated[
    str, Field(pattern=r"^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)")
]

MD5HashStr: TypeAlias = Annotated[str, Field(pattern=r"^([a-f\d]{32}|[A-F\d]{32})$")]



#!/usr/bin/env python

"""Updates the schemas for both the dev environment and the staging/production
environment.

To prepare the schemas for release first check the changes with the --prod flag:

    $ ./tools/update-schemas.py --prod --diff

This will compare the schemas with all `$id` and `url` fields URLs removed which
in theory can be present anywhere in any schema. If the output looks as expected
you can write the production schemas:

    $ ./tools/update-schemas.py --prod --force

"""

from __future__ import annotations

import argparse
import difflib
import json
import os
import subprocess
import sys
from copy import deepcopy
from enum import Enum, auto
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar

if TYPE_CHECKING:
    from fmu.datamodels._schema_base import SchemaBase

GREEN = "\033[32m"
RED = "\033[31m"
YELLOW = "\033[93m"
NC = "\033[0m"
BOLD = "\033[1m"
PASS = f"[{BOLD}{GREEN}✔{NC}]"
FAIL = f"[{BOLD}{RED}✖{NC}]"
INFO = f"[{BOLD}{YELLOW}+{NC}]"

T = TypeVar("T", dict, list, object)


def _get_parser() -> argparse.ArgumentParser:
    """Construct parser object."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d",
        "--diff",
        action="store_true",
        help="Show a diff between the current schema and the new one in output.",
    )
    parser.add_argument(
        "-p",
        "--prod",
        action="store_true",
        help="Produce schemas with production URLs",
    )
    parser.add_argument(
        "-t",
        "--test",
        action="store_true",
        help="Run as normal, but don't write the file.",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Force the script to overwrite the current schema with the new schema.",
    )
    parser.add_argument(
        "--force-contractual",
        action="store_true",
        dest="force_contractual",
        help="Force the script to accept a removal of a $contractual item.",
    )
    return parser


def _show_git_diff(output_filepath: Path) -> None:
    command = ["git", "diff", str(output_filepath)]
    print(INFO, f"running `{' '.join(command)}` ...")
    output = subprocess.run(command, check=False, capture_output=True, text=True)
    diff_str = "\n      ".join(output.stdout.split("\n"))
    print(f"      {diff_str}")  # To indent the first line too


def _show_py_diff(existing_schema: dict[str, Any], new_schema: dict[str, Any]) -> None:
    existing = json.dumps(existing_schema, indent=2, sort_keys=True)
    new = json.dumps(new_schema, indent=2, sort_keys=True)
    diff = difflib.unified_diff(
        existing.splitlines(),
        new.splitlines(),
        lineterm="",
        fromfile="existing schema",
        tofile="new schema",
    )
    diff_str = "\n      ".join(diff)
    print(f"      {diff_str}")


def _remove_schema_ids(schema: T) -> T:
    """Recursively remove all '$id' and 'url' fields from a schema."""
    if isinstance(schema, dict):
        return {
            key: _remove_schema_ids(value)
            for key, value in schema.items()
            if key not in ("$id", "url")
        }
    if isinstance(schema, list):
        return [_remove_schema_ids(item) for item in schema]
    return schema


def _schemas_are_equivalent(
    existing_schema: dict[str, Any],
    new_schema: dict[str, Any],
    is_release: bool,
    release_url: str,
) -> bool:
    """Checks that schemas are equivalent.

    If this is a release, it first removes all $id fields to ensure no other fields are
    being changed. It re-applies the root $id field (self-reference) to ensure that they
    do match.
    """
    if is_release:
        existing_schema = _remove_schema_ids(deepcopy(existing_schema))
        existing_schema["$id"] = release_url

        new_schema_id = new_schema["$id"]
        new_schema = _remove_schema_ids(deepcopy(new_schema))
        new_schema["$id"] = new_schema_id

    return existing_schema == new_schema


class SchemaUpdateResult(Enum):
    """The result of updating a schema."""

    UNCHANGED = auto()
    UPDATED = auto()
    FAILED = auto()


class SchemaWriter:
    """Handles schema finding, updating, and writing."""

    def __init__(
        self,
        *,
        is_release: bool = False,
        show_diff: bool = False,
        is_test: bool = False,
        force_overwrite: bool = False,
        force_contractual: bool = False,
    ) -> None:
        self.force_overwrite = force_overwrite
        self.is_test = is_test
        self.is_release = is_release
        self.show_diff = show_diff
        self.force_contractual = force_contractual

    def write_schema(
        self,
        schema_base: type[SchemaBase],
    ) -> SchemaUpdateResult:
        """Write schema to file after some checking."""
        output_filepath = self._get_output_filepath(schema_base.PATH)
        self._ensure_output_path(output_filepath.parent)

        new_schema = schema_base.dump()

        try:
            existing_schema = self._load_existing_schema(output_filepath)
        except FileNotFoundError:
            return self._write_new_schema(output_filepath, new_schema, schema_base)
        except json.JSONDecodeError as e:
            if not self.force_overwrite:
                print(FAIL, f"Invalid JSON in existing schema: {e}")
                return SchemaUpdateResult.FAILED
            return self._write_new_schema(output_filepath, new_schema, schema_base)

        return self._handle_existing_schema(
            existing_schema,
            new_schema,
            schema_base,
            output_filepath,
        )

    def contractual_item_removed(self, fmu_result_schema: type[SchemaBase]) -> bool:
        newest_existing_schema = self._load_existing_schema(
            self._find_newest_existing_schema_path(fmu_result_schema)
        )
        existing_contractual = newest_existing_schema["$contractual"]
        new_contractional = fmu_result_schema.dump()["$contractual"]
        if not set(existing_contractual).issubset(set(new_contractional)):
            print(
                FAIL,
                "🚨 A contractual item has been removed from the new schema. "
                "This is not supported, unless you force the removal by running the "
                "script with the '--force-contractual' argument.",
            )
            return True
        return False

    def _get_output_filepath(self, schema_path: Path) -> Path:
        """Returns a Path with the appropriate output location, without the filename."""
        project_root = Path(__file__).parent.parent.resolve()  # absolute path of ../../
        return project_root / schema_path

    def _ensure_output_path(self, dir_path: Path) -> None:
        dir_exists = dir_path.exists()
        if dir_exists and not dir_path.is_dir():
            raise ValueError(f"Path exists but is not a directory: {dir_path}")

        if not self.is_test:
            dir_path.mkdir(parents=True, exist_ok=True)
        if not dir_exists:
            print(INFO, f"Created directory: {dir_path}")

    def _load_existing_schema(self, filepath: Path) -> dict[str, Any]:
        if not filepath.exists():
            raise FileNotFoundError(f"Schema not found: {filepath}")

        with open(filepath, encoding="utf-8") as f:
            return json.load(f)

    def _find_newest_existing_schema_path(
        self, fmu_result_schema: type[SchemaBase]
    ) -> Path:
        output_filepath = self._get_output_filepath(fmu_result_schema.PATH)
        if output_filepath.exists():
            newest_existing_schema_path = output_filepath
        else:
            versions = [
                line.strip("# ")
                for line in fmu_result_schema.VERSION_CHANGELOG.splitlines()
                if line.strip().startswith("####")
            ]
            current_version = versions[0]
            previous_version = versions[1]
            previous_schema_path = str(output_filepath).replace(
                current_version, previous_version
            )
            newest_existing_schema_path = Path(previous_schema_path)
        return newest_existing_schema_path

    def _handle_existing_schema(
        self,
        existing_schema: dict[str, Any],
        new_schema: dict[str, Any],
        schema_base: type[SchemaBase],
        output_filepath: Path,
    ) -> SchemaUpdateResult:
        """Handles the case when a schema version we're writing to already exists.

        Typically this means we need a version bump. In rare cases it could mean we need
        to force an overwrite.
        """
        if not self.force_overwrite and not _schemas_are_equivalent(
            existing_schema, new_schema, self.is_release, schema_base.url()
        ):
            print(
                FAIL,
                f"🚨 {BOLD}{schema_base.FILENAME}{NC} version "
                f"{BOLD}{schema_base.VERSION}{NC} "
                "has changed. does it need a new version?",
            )
            if self.show_diff:
                _show_py_diff(existing_schema, new_schema)
            return SchemaUpdateResult.FAILED

        if (
            self.is_release
            and not self.force_overwrite
            and existing_schema["$id"] != new_schema["$id"]
        ):
            print(
                FAIL,
                f"🚨 {BOLD}{schema_base.FILENAME}{NC} version "
                f"{BOLD}{schema_base.VERSION}{NC}: mismatch in '$id'. "
                "you probably need to run `./tools/update-schemas.py --prod --force`",
            )
            if self.show_diff:
                _show_py_diff(existing_schema, new_schema)
            return SchemaUpdateResult.FAILED

        if new_schema == existing_schema:
            print(
                PASS,
                f"{BOLD}{schema_base.FILENAME}{NC} version "
                f"{BOLD}{schema_base.VERSION}{NC} unchanged",
            )
            return SchemaUpdateResult.UNCHANGED
        return self._write_new_schema(output_filepath, new_schema, schema_base)

    def _write_new_schema(
        self,
        output_filepath: Path,
        schema: dict[str, Any],
        schema_base: type[SchemaBase],
    ) -> SchemaUpdateResult:
        """Writes the schema to disk."""
        if not self.is_test:
            with open(output_filepath, "w", encoding="utf-8") as f:
                f.write(json.dumps(schema, indent=2, sort_keys=True))

        print(
            PASS,
            f"{BOLD}{schema_base.FILENAME}{NC} version {BOLD}{schema_base.VERSION}{NC} "
            f"written to:\n      {output_filepath}",
        )

        if self.show_diff:
            _show_git_diff(output_filepath)
        return SchemaUpdateResult.UPDATED


def main() -> None:
    parser = _get_parser()
    args = parser.parse_args()

    if args.force:
        print(INFO, "Forcing schema overwrite")

    os.environ["DEV_SCHEMA"] = "" if args.prod else "1"
    # Ensures URLs will differ based on above
    import fmu.datamodels as models  # noqa

    writer = SchemaWriter(
        is_release=args.prod,
        show_diff=args.diff,
        is_test=args.test,
        force_overwrite=args.force,
        force_contractual=args.force_contractual,
    )

    if not writer.force_contractual and writer.contractual_item_removed(
        models.FmuResultsSchema
    ):
        sys.exit(1)

    schema_update_results = []
    for schema in models.schemas:
        update_result = writer.write_schema(schema)
        schema_update_results.append(update_result)

    if SchemaUpdateResult.FAILED in schema_update_results:
        sys.exit(1)


if __name__ == "__main__":
    main()



from typing import Literal, get_args, get_origin

from fmu.datamodels.fmu_results import data, enums
from tests.utils import _get_pydantic_models_from_annotation


def test_all_attribute_enums_in_anyattributespecification() -> None:
    """
    Test that all attribute enums are represented with a model
    in AnyAttributeSpecification.
    """

    models = _get_pydantic_models_from_annotation(
        data.AnyAttributeSpecification.model_fields["root"].annotation
    )

    enums_in_anyattributespecification = []
    for model in models:
        # attribute is used as a discriminator in AnyAttributeSpecification and
        # should be present for all models
        assert "attribute" in model.model_fields
        attribute_annotation = model.model_fields["attribute"].annotation

        # check that the annotation is a Literal
        assert get_origin(attribute_annotation) == Literal

        # get_args will unpack the enum from the Literal
        # into a tuple, should only be one Literal value
        assert len(get_args(attribute_annotation)) == 1

        # the literal value should be an enum
        attribute_enum = get_args(attribute_annotation)[0]
        assert isinstance(attribute_enum, enums.PropertyAttribute)

        enums_in_anyattributespecification.append(attribute_enum)

    # finally check that all attribute enums are represented
    for attribute_enum in enums.PropertyAttribute:
        assert attribute_enum in enums_in_anyattributespecification

    # and that number of models in AnyAttributeSpecification matches
    # number of attribute enums
    assert len(models) == len(enums.PropertyAttribute)



from __future__ import annotations

from typing import Annotated, Literal

from pydantic import BaseModel, Field, RootModel

from . import enums


class AttributeSpecification(BaseModel):
    """Specifies a property attribute and its characteristics."""

    attribute: enums.PropertyAttribute
    """The name of the property attribute."""
    is_discrete: bool
    """Whether the property is discrete (categorical) or continuous."""
    min_value: float | None = None
    """Minimum value for the property, if applicable."""
    max_value: float | None = None
    """Maximum value for the property, if applicable."""


class BulkVolumeGasAttributeSpecification(AttributeSpecification):
    """Specifications related to the bulk volume gas attribute."""

    attribute: Literal[enums.PropertyAttribute.bulk_volume_gas]
    is_discrete: bool = False
    min_value: float = 0


class BulkVolumeOilAttributeSpecification(AttributeSpecification):
    """Specifications related to the bulk volume oil attribute."""

    attribute: Literal[enums.PropertyAttribute.bulk_volume_oil]
    is_discrete: bool = False
    min_value: float = 0


class FaciesAttributeSpecification(AttributeSpecification):
    """Specifications related to the facies attribute."""

    attribute: Literal[enums.PropertyAttribute.facies]
    is_discrete: bool = True
    min_value: float = 0


class FluidIndicatorAttributeSpecification(AttributeSpecification):
    """Specifications related to the fluid indicator attribute."""

    attribute: Literal[enums.PropertyAttribute.fluid_indicator]
    is_discrete: bool = True
    min_value: float = 0


class NetToGrossAttributeSpecification(AttributeSpecification):
    """Specifications related to the net-to-gross attribute."""

    attribute: Literal[enums.PropertyAttribute.net_to_gross]
    is_discrete: bool = True
    min_value: float = 0
    max_value: float = 1


class PermeabilityAttributeSpecification(AttributeSpecification):
    """Specifications related to the permeability attribute."""

    attribute: Literal[enums.PropertyAttribute.permeability]
    is_discrete: bool = False
    min_value: float = 0


class PermeabilityVerticalAttributeSpecification(AttributeSpecification):
    """Specifications related to the vertical permeability attribute."""

    attribute: Literal[enums.PropertyAttribute.permeability_vertical]
    is_discrete: bool = False
    min_value: float = 0


class PorosityAttributeSpecification(AttributeSpecification):
    """Specifications related to the porosity attribute."""

    attribute: Literal[enums.PropertyAttribute.porosity]
    is_discrete: bool = False
    min_value: float = 0
    max_value: float = 1


class RegionsAttributeSpecification(AttributeSpecification):
    """Specifications related to the regions attribute."""

    attribute: Literal[enums.PropertyAttribute.regions]
    is_discrete: bool = True
    min_value: float = 0


class SaturationGasAttributeSpecification(AttributeSpecification):
    """Specifications related to the gas saturation attribute."""

    attribute: Literal[enums.PropertyAttribute.saturation_gas]
    is_discrete: bool = False
    min_value: float = 0
    max_value: float = 1


class SaturationOilAttributeSpecification(AttributeSpecification):
    """Specifications related to the oil saturation attribute."""

    attribute: Literal[enums.PropertyAttribute.saturation_oil]
    is_discrete: bool = False
    min_value: float = 0
    max_value: float = 1


class SaturationWaterAttributeSpecification(AttributeSpecification):
    """Specifications related to the water saturation attribute."""

    attribute: Literal[enums.PropertyAttribute.saturation_water]
    is_discrete: bool = False
    min_value: float = 0
    max_value: float = 1


class VolumeShaleAttributeSpecification(AttributeSpecification):
    """Specifications related to the volume shale attribute."""

    attribute: Literal[enums.PropertyAttribute.volume_shale]
    is_discrete: bool = False
    min_value: float = 0
    max_value: float = 1


class ZonationAttributeSpecification(AttributeSpecification):
    """Specifications related to the zonation attribute."""

    attribute: Literal[enums.PropertyAttribute.zonation]
    is_discrete: bool = True
    min_value: float = 0


class AnyAttributeSpecification(RootModel):
    """
    Root model that allows for property attribute specifications with different
    attribute types to be placed within it. The discriminator field is ``attribute``,
    which indicates the specific property attribute being described.
    """

    root: Annotated[
        BulkVolumeGasAttributeSpecification
        | BulkVolumeOilAttributeSpecification
        | FaciesAttributeSpecification
        | FluidIndicatorAttributeSpecification
        | NetToGrossAttributeSpecification
        | PermeabilityAttributeSpecification
        | PermeabilityVerticalAttributeSpecification
        | PorosityAttributeSpecification
        | RegionsAttributeSpecification
        | SaturationGasAttributeSpecification
        | SaturationOilAttributeSpecification
        | SaturationWaterAttributeSpecification
        | VolumeShaleAttributeSpecification
        | ZonationAttributeSpecification,
        Field(discriminator="attribute"),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic import BaseModel, Field, RootModel

from fmu.datamodels._schema_base import FMU_SCHEMAS_PATH, SchemaBase

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any

    from fmu.datamodels.types import VersionStr


class FluidContactOutlineResultRow(BaseModel):
    """Represents the columns of a row in a fluid contact outline export.

    These fields are the current agreed upon standard result. Changes to the fields or
    their validation should cause the version defined in the standard result schema to
    increase the version number in a way that corresponds to the schema versioning
    specification (i.e. they are a patch, minor, or major change)."""

    X_UTME: float
    """The X coordinate this row represents. Required."""

    Y_UTMN: float
    """The Y coordinate this row represents. Required."""

    Z_TVDSS: float
    """The Z coordinate (depth) this row represents. Required."""

    POLY_ID: int = Field(ge=0)
    """Index column. The id of the polygon which this row represents. Required."""


class FluidContactOutlineResult(RootModel):
    """Represents the resultant fluid contact outline parquet file, which is
    naturally a list of rows.

    Consumers who retrieve this parquet file must read it into a json-dictionary
    equivalent format to validate it against the schema."""

    root: list[FluidContactOutlineResultRow]


class FluidContactOutlineSchema(SchemaBase):
    """This class represents the schema that is used to validate the fault lines
    table being exported. This means that the version, schema filename, and schema
    location corresponds directly with the values and their validation constraints,
    documented above."""

    VERSION: VersionStr = "0.1.0"
    """The version of this schema."""

    VERSION_CHANGELOG: str = """
    #### 0.1.0

    This is the initial schema version.
    """

    FILENAME: str = "fluid_contact_outline.json"
    """The filename this schema is written to."""

    PATH: Path = FMU_SCHEMAS_PATH / "file_formats" / VERSION / FILENAME
    """The local and URL path of this schema."""

    @classmethod
    def dump(cls) -> dict[str, Any]:
        return FluidContactOutlineResult.model_json_schema(
            schema_generator=cls.default_generator()
        )



from .access import Access, Asset, Ssdl, SsdlAccess
from .masterdata import (
    CoordinateSystem,
    CountryItem,
    DiscoveryItem,
    FieldItem,
    Masterdata,
    Smda,
    StratigraphicColumn,
)
from .tracklog import (
    OperatingSystem,
    SystemInformation,
    Tracklog,
    TracklogEvent,
    TracklogSource,
    User,
    Version,
)

__all__ = [
    "Access",
    "SsdlAccess",
    "Asset",
    "Ssdl",
    "Masterdata",
    "Smda",
    "StratigraphicColumn",
    "CoordinateSystem",
    "CountryItem",
    "FieldItem",
    "DiscoveryItem",
    "Tracklog",
    "OperatingSystem",
    "SystemInformation",
    "TracklogEvent",
    "TracklogSource",
    "User",
    "Version",
]



"""Data mapping models between systems."""

from collections.abc import Iterator
from enum import StrEnum
from typing import Annotated, Literal, Self
from uuid import UUID

from pydantic import BaseModel, Field, RootModel, field_validator, model_validator


class MappingType(StrEnum):
    """The discriminator used between mapping types.

    Each of these types should have their own mapping class derived from a base
    mapping type, e.g. IdentifierMapping.
    """

    stratigraphy = "stratigraphy"


class RelationType(StrEnum):
    """The kind of relation this mapping represents."""

    primary = "primary"
    """The primary unofficial identifier."""

    alias = "alias"
    """Alias of a primary unofficial identifier."""

    equivalent = "equivalent"
    """A name used in the source system that is the same as the official name.

    For example, if an RMS stratigraphic name is the same as the SMDA name."""


class DataSystem(StrEnum):
    """The system or application data is being mapping to or from."""

    rms = "rms"
    smda = "smda"
    fmu = "fmu"


class BaseMapping(BaseModel):
    """The base mapping containing the fields all mappings should contain."""

    source_system: DataSystem
    target_system: DataSystem
    mapping_type: MappingType
    relation_type: RelationType

    @model_validator(mode="after")
    def validate_systems_differ(self) -> "BaseMapping":
        """Ensure source and target systems are different.

        Mappings between FMU sources are allowed.
        """

        if self.source_system == self.target_system:
            # Allow FMU to map to FMU.
            if self.source_system == DataSystem.fmu:
                return self

            raise ValueError(
                f"source_system and target_system must differ, "
                f"both are '{self.source_system}'"
            )
        return self


class IdentifierMapping(BaseMapping):
    """Base class for a one-to-one mapping of identifiers.

    This mapping represents an identifier from one source and correlates it to an
    identifier in a target. Most often this target will be some official masterdata
    store like SMDA.
    """

    source_id: str
    source_uuid: UUID | None = None
    target_id: str
    target_uuid: UUID | None = None

    @field_validator("source_id", "target_id")
    def validate_ids_not_empty(cls: Self, v: str) -> str:
        """Ensure IDs are not empty strings."""
        if not v or not v.strip():
            raise ValueError("An identifier cannot be an empty string")
        return v.strip()

    @model_validator(mode="after")
    def validate_equivalent_relation(self) -> Self:
        """Ensure relation_type=equivalent is only used when source and target match."""
        if (
            self.relation_type == RelationType.equivalent
            and self.source_id != self.target_id
        ):
            raise ValueError(
                "Equivalent mapping requires matching source_id/target_id; "
                f"got source_id='{self.source_id}', target_id='{self.target_id}'"
            )
        return self


class StratigraphyIdentifierMapping(IdentifierMapping):
    """Represents a stratigraphy mapping.

    This is a mapping from stratigraphic identifiers (tops, zones, etc.) to official
    identifiers in SMDA.
    """

    mapping_type: Literal[MappingType.stratigraphy] = MappingType.stratigraphy


AnyIdentifierMapping = Annotated[
    StratigraphyIdentifierMapping, Field(discriminator="mapping_type")
]


class StratigraphyMappings(RootModel[list[StratigraphyIdentifierMapping]]):
    """Collection of all stratigraphy mappings."""

    root: list[StratigraphyIdentifierMapping]

    def __getitem__(self: Self, index: int) -> StratigraphyIdentifierMapping:
        """Retrieves a stratigraphy mapping from the list using the specified index."""
        return self.root[index]

    def __iter__(  # type: ignore[override]
        self: Self,
    ) -> Iterator[StratigraphyIdentifierMapping]:
        """Returns an iterator for the stratigraphy mappings."""
        return iter(self.root)

    def __len__(self: Self) -> int:
        """Returns the number of stratigraphy mappings."""
        return len(self.root)

    def get_by_source(
        self, source_id: str, source_system: DataSystem = DataSystem.rms
    ) -> list[StratigraphyIdentifierMapping]:
        """Get all stratigraphy mappings from a source identifier."""
        return [
            m
            for m in self.root
            if m.source_id == source_id and m.source_system == source_system
        ]

    def get_by_target(
        self,
        target_id: str,
        target_system: DataSystem = DataSystem.smda,
    ) -> list[StratigraphyIdentifierMapping]:
        """Get all stratigraphy mappings to a target identifier."""
        return [
            m
            for m in self.root
            if m.target_id == target_id and m.target_system == target_system
        ]

    def get_official_name(
        self,
        rms_name: str,
    ) -> str | None:
        """Get the official SMDA name for an RMS stratigraphy identifier.

        Args:
            rms_name: The RMS name

        Returns:
            The official SMDA name, or None if not found
        """
        mappings = [
            m
            for m in self.root
            if m.source_id == rms_name
            and m.source_system == DataSystem.rms
            and m.target_system == DataSystem.smda
        ]
        return mappings[0].target_id if mappings else None



import pyarrow as pa
import pytest

from fmu.datamodels.standard_results.ert_parameters import (
    ConstParameter,
    DerrfParameter,
    DUnifParameter,
    ErrfParameter,
    ErtDistribution,
    ErtParameterMetadata,
    LogNormalParameter,
    LogUnifParameter,
    NormalParameter,
    RawParameter,
    TriangularParameter,
    TruncatedNormalParameter,
    UniformParameter,
)


def test_to_pa_metadata_serialization() -> None:
    """Dumping Parquet schemas serializes to dict[bytes, bytes].

    Uses DUnif because it has both float and int."""
    dist = DUnifParameter(
        distribution=ErtDistribution.dunif,
        min=0.0,
        max=1.0,
        steps=100,
        group="GLOBVAR",
        input_source="sampled",
    )
    assert dist.to_pa_metadata() == {
        b"group": b'"GLOBVAR"',
        b"input_source": b'"sampled"',
        b"distribution": b'"dunif"',
        b"min": b"0.0",
        b"max": b"1.0",
        b"steps": b"100",
    }

    field = pa.field("PARAM1", pa.float64(), metadata=dist.to_pa_metadata())
    assert field.metadata == {
        b"group": b'"GLOBVAR"',
        b"input_source": b'"sampled"',
        b"distribution": b'"dunif"',
        b"min": b"0.0",
        b"max": b"1.0",
        b"steps": b"100",
    }


@pytest.mark.parametrize(
    "parameter_class, value_dict",
    [
        (UniformParameter, {"distribution": "uniform", "min": 0.0, "max": 1.0}),
        (LogUnifParameter, {"distribution": "logunif", "min": 0.0, "max": 1.0}),
        (NormalParameter, {"distribution": "normal", "mean": 0.2, "std": 0.1}),
        (
            TruncatedNormalParameter,
            {
                "distribution": "truncated_normal",
                "min": 0.0,
                "max": 1.0,
                "mean": 0.2,
                "std": 0.1,
            },
        ),
        (LogNormalParameter, {"distribution": "lognormal", "mean": 0.2, "std": 0.1}),
        (RawParameter, {"distribution": "raw"}),
        (ConstParameter, {"distribution": "const", "value": 2.0}),
        (
            TriangularParameter,
            {"distribution": "triangular", "min": 0.0, "max": 3.0, "mode": 2.0},
        ),
        (
            ErrfParameter,
            {
                "distribution": "errf",
                "min": 0.0,
                "max": 3.0,
                "skewness": 2.0,
                "width": 10.0,
            },
        ),
        (
            DerrfParameter,
            {
                "distribution": "derrf",
                "min": 0.0,
                "max": 3.0,
                "skewness": 2.0,
                "width": 10.0,
                "steps": 1000.0,
            },
        ),
        (
            DUnifParameter,
            {"distribution": "dunif", "min": 0.0, "max": 1.0, "steps": 100},
        ),
    ],
)
def test_from_pa_metadata_roundtrip(
    parameter_class: ErtParameterMetadata, value_dict: dict[str, int | float]
) -> None:
    """Roundtrip serialization-deserialization works for all parameters."""
    input_source = (
        "design_matrix" if type(parameter_class) is RawParameter else "sampled"
    )
    dist = parameter_class.model_validate(
        {"group": "GLOBVAR", "input_source": input_source, **value_dict}
    )
    field = pa.field("PARAM1", pa.float64(), metadata=dist.to_pa_metadata())
    assert field.metadata is not None
    from_dist = parameter_class.from_pa_metadata(field.metadata)
    assert dist == from_dist



from __future__ import annotations

from pydantic import (
    BaseModel,
    Field,
)

from . import enums


class Access(BaseModel):
    """
    The ``access`` block contains information related to access control for
    this data object.
    """

    asset: Asset
    """A block containing information about the owner asset of these data.
    See :class:`Asset`."""

    classification: enums.Classification | None = Field(default=None)
    """The access classification level. See :class:`enums.Classification`."""


class SsdlAccess(Access):
    """
    The ``access`` block contains information related to access control for
    this data object, with legacy SSDL settings.
    """

    ssdl: Ssdl
    """A block containing information related to SSDL. See :class:`Ssdl`."""


class Asset(BaseModel):
    """The ``access.asset`` block contains information about the owner asset of
    these data."""

    name: str = Field(examples=["Drogon"])
    """A string referring to a known asset name."""


class Ssdl(BaseModel):
    """
    The ``access.ssdl`` block contains information related to SSDL.
    Note that this is kept due to legacy.
    """

    access_level: enums.Classification
    """The SSDL access level. See :class:`Classification`."""

    rep_include: bool
    """Flag if this data is to be shown in REP or not."""
